#!/usr/local/bin/perl -w
#
# SolServer is a network service that allows remote applications to run
# Hypoinverse as the standard earthquake location program. The remote
# client passes phase arrival information to the server over a TCP socket
# connection in Hypoinverse .arc format and receives the solution in the
# various formats over the same connection.
#
# The SolServer can handle multiple concurrent connections.
#
# HINTS
# * You can test a solserver using 'hypolocate'
#   Ex: % hypolocate -u amosite -p 6504 14419992
# * You can test a solserver interactively via telnet.
#   Ex: % telnet amosite 6504
# * If there's a firewall in place you will need to allow the ports used.
#
# SolServer processes are started and checked by startAllSolservers.pl. 
# This also creates a directory for each port service and sets up default configs.
#
#    usage: solserver [-p <port>] [-i <instFile>]
#
#     -p        : port number (default = $port)
#     -i        : Hypoinverse instruction file (default = "$DEFAULT_instFile")
#     -h        : print usage info
#
# DDG 8/14/08
#
# Assumptions:
# * This is running in a directory that contains Hypo config files
#   (As is created automatically by startAllSolservers.pl)
# * Actual data and results files will be in a subdir called ./data
# * ARC format input file will be of the form ./data/z<evid>_<child#>.arcin
# * ARC format output file will be of the form ./data/z<evid_<child#>>.arcout
# * PRT format output file will be of the form ./data/z<evid_<child#>>.prt
# * Hypoinverse instruction file may be spec'ed with -i on command line (defaults to "hypinst.")
#
# The "z" on the working files just makes it simpler to ls and delete them if
# there are any other files mingled in.
#
# DDG 2/19/2009 =======================================================================
# Added hypinst. file management commands. Menlo wanted to be able to change Hypoinverse
# setup from Jiggle for offshore events.
#
# |GetInstName| - returns the name of the current instruction file
# |GetInstFileList| - return list of available instruction files, one per line
# |GetInstFile| <filename> - return contents of this instruction file
# |GetInstFile| - return contents of currently set instruction file
# |PutInstFile| <filename> - accepts text until |EOT| or |EOF| and writes this instruction file
# |SetInstFile| <filename> - use this instruction file for the duration of the connection
# |SetInstFile| - set instruction file back to default
#
# Default instruction file: is "hypinst." in its working directory unless  it is
# be specified at startup with the "-i" command line switch. The contents of the
# default instruction file cannot be altered by clients.
#
# Custom instruction files: clients can create and modify their own custom 
# instruction files. Custom instruction files are stored persistently in the directory
# "<TPP_HOME>/solserver/hypinstFiles" and can be named anything except "hypinst.".
# To use a custom instruction file when running Hypoinverse a client must specify 
# the file with the |SetInstFile| command before locating an event. 
# The file will be used for the duration of that solserver connection.
#
# Things to remember:
# * the name "hypinst." is reserved for the default file and which may not be modified.
# * each solserver instance (port number) can have a different default file as specified with the "-i" command line switch.
# * all users sharing a solserver host will have access to all the custom instruction files.
# * all users can modify these shared files so you can't be sure a file hasn't been 
#    changed since you used it last.
# * Multiple users can overwrite files (use the same filename)
# * This is a consequence of allowing users to share files and making the files
#   accessible from different solserver port numbers.
#
# DDG 11/25/09 - added |EOT| to every response at request of AWW.

use strict;
use IO::Socket;
use Sys::Hostname;

my $debug = 1;

my $modelFile   = "";
my $stationFile = "";
my $port        = 6510;
my $localhost   = hostname();
my $childNo     = 0;

# pseudo-statics (globals)
our $ARCIN_FILE  = 0;
our $ARCOUT_FILE = 1;
our $PRT_FILE    = 2;
our $outFlag     = $ARCOUT_FILE;    # default

# Share this amoung subs
our $FilePrefix = "";

# parse optional switches (N.B. this 'shift's the @ARGV array)
use Getopt::Std;
use vars qw/ $opt_h $opt_i $opt_p /;    #prevent "only used once" error

getopts('hi:p:');                       # "d:" means a value must follow
if ($opt_h) { &usage(); exit; }

if ($opt_p) { $port = $opt_p; }

my $DEFAULT_instFile = "hypinst.";  # special filename - user can't edit this file
my $instPath = "../hypinstFiles/";
if ($opt_i) { $DEFAULT_instFile = $opt_i; }
my $instFile    = "";

# set directory paths and CD to the proper dir
my $root       = "$ENV{TPP_HOME}/solserver";
my $workingdir = "$root/${localhost}_${port}";
chdir $workingdir or die "Can't chdir to $workingdir: $!\n";

# Fail if default instruction file doesn't exist
if (&setInstFile($DEFAULT_instFile) == 0) {
    die "!! Terminating: $instFile does not exist.\n";
}
&parseInstFile($instFile);

# ignore SIGCHLD to prevent zombie child processes
$SIG{CHLD} = 'IGNORE';

# force print buffer to autoflush, else debug messages are delayed.
$| = 1;

my $time = localtime();
print "Solserver starting on $time\n";
print "    Hypoinverse instruction file = $instFile\n";
print "    Listening on port $port\n";

# TCP, max 5 clients can be queued
my $socket = new IO::Socket::INET(
				  LocalHost => '0.0.0.0',
				  LocalPort => $port,
				  Proto     => 'tcp',
				  Reuse     => 1,
				  Listen    => 5
				  )
    or die "Socket listener creation failed\n $!";

my $buf;
my $new_sock;
my $pid;

# #################################################################
LISTENER_LOOP:
    while ( $new_sock = $socket->accept() ) {

	# get info about who connected
	my $remote = getpeername($new_sock);
	my ( $rport, $riaddr ) = unpack_sockaddr_in($remote);
	my $rip   = inet_ntoa($riaddr);
	my $rhost = gethostbyaddr( $riaddr, AF_INET );
	
	$childNo++;
	print("Accepted client connecton: port= $rport  remote host= $rhost ($rip)\n");
	print("Forking child # $childNo\n");
	    
	    # Fork a new process to handle connection, current variable values are copied
	    $pid = fork();
	
	# Child Process starts running here =====
	if ( $pid == 0 ) {
	    
	    $socket->close();    # listener socket is of no use to child
	    &handleConnection( $new_sock, $childNo );

	}
	else {

	    $new_sock->close();    #else child can't fully close connection
	    next LISTENER_LOOP;
	}
    }    # end of "while"

# #################################################################

# should never get here
print "Parent exiting. Last child was $childNo\n";
close($socket);
exit;

# -----------------------------------------------------------------
# This does the heavy lifting of servicing the connection.
# Stays connected until it sees the |EOT| (end of transmission) tag or
# it does a |FILE|-|EOF| operation.
#
# Legal tags:
# (beginning tags are NOT case sensitive, ending tags are)
#
# -- Get config info --
# |GETINSTFILE| - return the contents of the Hypoinverse instruction file
# |GETSTANAME| - return the name of the Hypoinverse station file
# |GETSTAFILE| - return the contents of the Hypoinverse station file
# |GETCRHNAME| - return the name of the Hypoinverse crustal model file
# |GETCRUFILE| - return the contents of the Hypoinverse crustal model file
# -- Instruction Files --
# |GetInstName| - returns the name of the current instruction file
# |GetInstFileList| - return list of available instruction files, one per line
# |GetInstFile| <filename> - return contents of this instruction file
# |PutInstFile| <filename> - accepts text until |EOT| or |EOF|, writes to instruction file
# |SetInstFile| <filename> - use this instruction file for the duration of the connection
# -- Locate and Event --
# Style 1 (older style - closes connection)
# |FILE| <evid> = Beginning of the .arc format input
# |EOF|  = End-Of-File, locate, return result in ARC format and close socket.
#          Finishes by closing socket connection (for backward compatibility)
# Style 2 (newer style - does NOT close connection)
# |FILE| <evid> = Beginning of the .arc format input
# |LOC| = Locate the event (run hypoinverse). No results returned, you ask for them later.
#         Sends "|EOT|" to signal it's done. Socket remains open.
# -- After a Location --
# |PRT|  = return the PRT formatted results (from last event this session)
# |ARC|  = return the ARC format results (from last event this session)
# |EOT|  = end of transmission, returns: disconnect the socket connection and exit the child

sub handleConnection ($) {

    my ( $socket, $childNo ) = @_;

    $socket->autoflush(1);

    my @arcdata;
    my $evid    = 0;
    my $filename ="";
    my $tmpStr = "";

    print("<$childNo> Forked: I'm child # $childNo\n");

	# read the input form the socket until it signals its done
      READ_MODE:
	while (<$socket>) {
	    
	    print ":$_";          # just echo input to log
	    
# parse line of form "|command| <arg>"
	    my ($blank, $cmd, $arg) = split (/\|/, $_);

	    $cmd = uc($cmd);  # force to uppercase
	    $arg = &trim($arg);

	    if ( $cmd eq "EOT" ) {
		print "<$childNo> Child exiting.\n";
		close($socket);
		exit;                # Child process exits

		# One-shot commands, one line returns (no |EOT|)

	    }
	    elsif ( $cmd eq "GETCRHNAME" ) {
		print $socket $modelFile . "\n";
   print $socket "|EOT|\n";    # End Of Data tag
	    }
	    elsif ( $cmd eq "GETSTANAME" ) {
		print $socket $stationFile . "\n";
		print $socket "|EOT|\n";    # End Of Data tag
	    }
	    elsif ( $cmd eq "GETINSTNAME" ) {
		print $socket &getInstFile() . "\n";
		print $socket "|EOT|\n";    # End Of Data tag
	    }
	    elsif ( $cmd eq "SETINSTFILE" ) {
		if ($arg) {
		    &setInstFile($arg);
		} else {
		    &setInstFile($DEFAULT_instFile);
		}
	    }
	    elsif ( $cmd eq "GETCRHFILE" ) {
		&writeFileToSocketEOT( $modelFile, $socket );
	    }
	    elsif ( $cmd eq "GETSTAFILE" ) {
		&writeFileToSocketEOT( $stationFile, $socket );
	    }
	    # Write a particular INSTS file to the socket
	    elsif ( $cmd eq "GETINSTFILELIST" ) {
		my $str = $DEFAULT_instFile."\n";     # add default file
		$str .= `ls -1 $instPath`;            # add custom files
		$str .= "\|EOT\|\n";
		print $socket $str;
	    }
	    elsif ( $cmd eq "GETINSTFILE" ) {

		if ($arg) {$filename = $instPath.$arg} 
		else {$filename = &getInstFile()};

		&writeFileToSocketEOT( $filename, $socket );

		# Read a particular INSTS file from the socket & write it to disk
	    }
	    elsif ( $cmd eq "PUTINSTFILE" ) {
		$tmpStr = &trim(substr( $_, 13 ));      #trim whitespace
		if ($tmpStr) {$filename = $instPath.$tmpStr} 
		else {$filename = $instPath.$instFile};

		(my $endToken, @arcdata) = &readFromSocket($socket);

# write the file
		my $FH;
		open( $FH, ">$filename" );
		print $FH @arcdata;    # assumes blanks removed as $arcdata was created
		close $FH;

		print ":Wrote $filename\n";

		# Locate an event
	    }  elsif ( $cmd eq "FILE" ) {    #Example:  "|FILE| 8327789"
		$evid = substr( $_, 6 );
		chomp $evid;
		$evid =~ tr/0-9//cd; #strip out anything that's not numeric (like whitespace)
		undef @arcdata; # Start out with empty array

		my $endToken = "";

		if ($debug) { print "<$childNo> evid = :$evid:\n" }

		# read from the socket until you see an end token
		($endToken, @arcdata) = &readFromSocket($socket);

		if ($debug) { print "<$childNo> Exiting readFromSocket() .\n" }

		# Do the right thing for the given end token
		if ( $endToken =~ /^\|EOF\|/ or $endToken =~ /^\|LOC\|/ ) {

		    if ($debug) { print "<$childNo> Saw EOF.\n" }

		    &runHypo( $evid, @arcdata );

		    if ( $endToken =~ /^\|EOF\|/ ) {    # send results and close socket

			&sendResults( $evid, $socket );
			&writeFileToSocket( "${FilePrefix}.arcout", $socket );

			$socket->close();           # bail - signals client we're all done
			exit;

		    } else {                     # must be |LOC| - just signal that we're done
			print $socket "|EOT|\n";    # End Of Data tag
			next READ_MODE;             #break out of inner loop
		    }

		}
		
	    }  # dump results files

	    elsif ( $cmd eq "PRT" ) {
		&writeFileToSocketEOT( "${FilePrefix}.prt", $socket );
	    }
	    elsif ( $cmd eq "ARCOUT" ) {
		&writeFileToSocketEOT( "${FilePrefix}.arcout", $socket );
	    }
	    elsif ( $cmd eq "ARCIN" ) {
		&writeFileToSocketEOT( "${FilePrefix}.arcin", $socket );
	    }
	    else {
		$_ =~ s/(\n|\r)//g;   # chomp only removes \n not <cr>
		print $socket "?\n";
		print "Nonsense: /$_/\n";
	    }
	}    # end of while (<$socket>)
}

# =================================================================
# my $DEFAULT_instFile    = "hypinst.";  # special filename - user can't edit this file
# my $instFile    = $DEFAULT_instFile;
# my $instPath = "../hypinstFiles/";

# Enforce rule that $DEFAULT_instFile can't be changed and is in different place
sub setInstFile ($) {

    my ( $filename ) = @_;

    my $str = &getInstPath($filename);

    print "dbg: $str\n";

    if (-e $str) {       #file exists
	$instFile = $filename;
	return 1;
    }
    return 0;
}

sub getInstFile () {

    return &getInstPath($instFile);

}
# Get the path.filename relative to the working dir
# Ex: ".hypinst", "../hypoinstFiles/testinst.", etc.
sub getInstPath ($) {
    my ( $filename ) = @_;

    if ($filename eq $DEFAULT_instFile) {
	return $filename;
    }

    return $instPath.$filename;

}
# =================================================================
#
# Syntax: runHypo ($evid, $arcdata, $outFlag)
#
# Arcin format is fussy!
#
# * Must have exactly one blank line at start
# * Station lines must be padded with blanks out to 113
# * Summary line must be padded to 168

sub runHypo ($) {

    my ( $evid, @arcdata ) = @_;

# Set the starting part of all file names. It is globally scoped.
# Insure unique file names for this thread so multiple threads processing the same
# event don't step on one another's files

    $FilePrefix = "data/z${evid}_${childNo}";
    my $arcinFile = "${FilePrefix}.arcin";

    if ($debug) { print "<$childNo> Dumping to $arcinFile...\n" }

    # write the file
    open( FH, ">$arcinFile" );
    print "sum len: ", length($arcdata[0]), "\n" if ($debug);
    if (length($arcdata[0]) > 121) {
	print FH shift(@arcdata);
    }
    else {
	print FH "\n";  # BEWARE!: the .arcin file must have exactly one summary line at the top
    }
    print FH join "", @arcdata; # assumes blanks removed as @arcdata was created
    close FH;

    # Locate: run Hypoinverse
    if ($debug) { print "<$childNo> Running location for $evid \n" }

    &locate($evid);

}

# =================================================================
#
# Encapulates the format of working files which is: "data/z${evid}_${childNo}
#
sub getFilePrefix ($) {

    my ( $_evid, $_childNo ) = @_;

    return "data/z${_evid}_${_childNo}";

}

# =================================================================
# send the results back to the client
sub sendResults ($) {

    my ( $evid, $socket ) = @_;

    # Set output file type based on how flag was set earlier
    my $fileType = "arcout";    #default

    if ( $outFlag == $PRT_FILE ) {
	$fileType = "prt";
    }
    elsif ( $outFlag == $ARCOUT_FILE ) {
	$fileType = "arcout";
    }
    elsif ( $outFlag == $ARCIN_FILE ) {
	$fileType = "arcin";
    }

    my $filename = "${FilePrefix}.${fileType}";

    &writeFileToSocket( $filename, $socket );

    if ($debug) { print "<$childNo> Wrote $filename to socket.\n"; }

}

# =================================================================
# Parse some important stuff out of the Hypoinverse instruction file
# STA - station list
# CRU - crustal file
#
# Assumes a line like "CRH 1 '../hadley.cru'           /crustal model (Hadley/Kanamori)"
# Where the filename is bounded by single quotes.

sub parseInstFile ($) {
    my ($inst_file) = @_;

    # suck the file into an array
    open( my $FILE, $inst_file ) || die "Can't open file: $inst_file\n";
    my @list = <$FILE>;

    my @str = ( grep ( /CRH/, @list ) );
    my $knt = @str;
    if ( $knt > 0 ) {
	my @array = split( /\'/, $str[ $knt - 1 ] );
	$modelFile = $array[1];
    }
    
    @str = ( grep ( /STA/, @list ) );
    $knt = @str;
    if ( $knt > 0 ) {
	my @array = split( /\'/, $str[ $knt - 1 ] );
	$stationFile = $array[1];
    }

    close $FILE;

    # print " $modelFile   $stationFile \n";
}

# -----------------------------------------------------------------
# write the given file to the socket
# Syntax: writeFileToSocket (file, socket)
sub writeFileToSocket ($) {

    my ( $file, $socket ) = @_;

    if ( open( my $FILE, $file ) ) {

	if ($debug) { print "<$childNo> Writing $file to socket...\n" }

	while (<$FILE>) {
	    print $socket "$_";
	}

	close($FILE);

    }
    else {
	print "No such file: $file\n";
	print $socket "No such file: $file\n";
    }

}

# -----------------------------------------------------------------
# write the given file to the socket
# Syntax: writeFileToSocket (file, socket)
sub writeFileToSocketEOT ($) {

    my ( $file, $socket ) = @_;

    if ( open( my $FILE, $file ) ) {

	if ($debug) { print "<$childNo> Writing $file to socket...\n" }

	while (<$FILE>) {
	    print $socket "$_";
	}

	close($FILE);
	print $socket "|EOT|\n";

    }
    else {
	print "No such file: $file\n";
	print $socket "No such file: $file\n";
    }

}

# -----------------------------------------------------------------
# Read from socket and accumulate in a buffer until you see an ending
# token. Returns the ending token string at the end.
# If we terminate due to a timeout return "|TIMEOUT|"
# Does not close the socket.

sub readFromSocket {

    my ( $socket ) = @_;

    my @buffer;

    while (<$socket>) {

	if ($debug) { print "<$childNo> " . $_; }

	if ( $_ =~ /^\|EOF\|/ or $_ =~ /^\|LOC\|/ or $_ =~ /^\|EOT\|/ ) {
	    return ($_, @buffer);   # return the terminator and the buffer

	} else {

	    # accumulate data

	    next if /^(\s)*$/;           # skip blank lines

	    push @buffer, $_;
	}
    }
}
# -----------------------------------------------------------------
# Locate the event
# Syntax: locate (evid)
sub locate ($) {
    my ($evid) = @_;

    my $hypoExe = "$root/bin/hyp2000";

    my $cmdFile = "${FilePrefix}.cmd";

    open( my $FILE, ">$cmdFile" || die "Can't create file: $cmdFile\n $!" );

    # NOTE: the main Hypoinverse behavior is defined in the $instFile
    #       Any universal modifications are added below.
    #       Currently only event-specific input/output files are so defined.
    # Write event-specific stuff to a command file
    my $inst = &getInstFile();
    print $FILE <<"EOF";
\@${inst}
PHS '${FilePrefix}.arcin'
ARC '${FilePrefix}.arcout'
PRT '${FilePrefix}.prt'
LOC
STO
EOF

	# run Hypoinverse with event-specific command file as input
	print `$hypoExe < $cmdFile`;

}
# -----------------------------------------------------------------
# Perl trim function to remove whitespace from the start and end of the string
sub trim($)
{
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

# -----------------------------------------------------------------
sub usage() {
    print <<"EOF";

Run the SolServer service

usage: $0 [-p $port] [-i $instFile]

-p        : port number (default = $port)
-i        : Hypoinverse instruction file (default = "$instFile")
-h        : print usage info

example: $0 -p 6610 -i hypo.inst

EOF
    exit;
}
